// RegX -- Regular Expression Matching Class

// This Regular Expression Matching module is an Object-based one
// derived from Rob Pike's original SAM editor functions.
// The text strings (both pattern and text to be matched) must be
// passed as 'Cstrips' -- objects that feed each sequential character
// on demand as an *integer* value (allowing for 16-bit unicode);
// end-of-text is signalled by -1 (errors by other negative values).

// Boolean options may be set (directly in the RegX object):
//	(false by default except eotiseol)
//		shortest	-- find shortest rather than longest match.
//		nlisany		-- 'ANY' ('.') matches newline as well
//		eotiseol	-- set false to prevent end-of-text (without
//					newline) being taken as end-of-line.
//		nocase		-- make matches case-insensitive.


////////////////////////////////////////////////////////////////////////
//	This code is heavily derived [read 'largely lifted'...] from
//	the regex code in Rob Pike's 'Sam' editor.  That original code
//	has the following Copyright notice:
/**********************************************************************
 * The authors of this software are Rob Pike and Howard Trickey.
 *		Copyright (c) 1998 by Lucent Technologies.
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR LUCENT TECHNOLOGIES MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************************/

//	The modifications herein are Copyright (c) 1999-2003 by Peter J. Goodeve
//	-- All Rights Reserved.  However, the conditions of the above notice
//	are considered to apply to the modified code as well, with the
//	added condition that this extension notice also be included in
//	any distribution.
////////////////////////////////////////////////////////////////////////

// Vers: 2005:03:2
// 'nlisany' now affects negated character class as well
// -- more special chars like '\r'

#include "regx.h"
#include "cstrip.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

// #define RXDEBUG

#ifdef	RXDEBUG
 #define DPRINTF(s) printf s
 char * iname(int t);
#else
 #define DPRINTF(s)
#endif

// for temporary quick-checks...
// Always on:
#define XDPRINTF(s) printf s
// Always off:
#define ZDPRINTF(s)

/*
 * State Machine Program Instruction
 */
struct Inst
{
	long	type;	/* < 0x10000 ==> literal, otherwise action */
	union {
		int sid;
		int subid;
		int ricode;
		struct Inst *other;
		struct Inst *right;
	};
	union{
		struct Inst *left;
		struct Inst *next;
	};
};

struct Ilist
{
	Inst	*inst;		/* Instruction of the thread */
	Rangeset se;
	Posn	startp;		/* first char of match */
};

#define	NLIST	128

/*
 * Actions and Tokens
 *
 *	0x100xx are operators, value == precedence
 *	0x200xx are tokens, i.e. operands for operators
 */
#define	OPERATOR	0x10000	/* Bitmask of all operators */
#define	START		0x10000	/* Start, used for marker on stack */
#define	RBRA		0x10001	/* Right bracket, ) */
#define	LBRA		0x10002	/* Left bracket, ( */
#define	OR		0x10003	/* Alternation, | */
#define	CAT		0x10004	/* Concatentation, implicit operator */
#define	STAR		0x10005	/* Closure, * */
#define	PLUS		0x10006	/* a+ == aa* */
#define	QUEST		0x10007	/* a? == a|nothing, i.e. 0 or 1 a's */
#define	ANY		0x20000	/* Any character but newline, . */
#define	NOP		0x20001	/* No operation, internal use only */
#define	BOL		0x20002	/* Beginning of line, ^ */
#define	EOL		0x20003	/* End of line, $ */
#define	CICODE		0x20004	/* Character class, [] */
#define	NCICODE		0x20005	/* Negated character class, [^] */
#define	END		0x20077	/* Terminate: match found */

#define	ISATOR		0x10000
#define	ISAND		0x20000

/*
 * Parser Information
 */
struct Node
{
	Inst	*first;
	Inst	*last;
};

struct _RegX_Internal {
	_RegX_Internal(RegX *r);
	~_RegX_Internal();
	#define	NPROG	1024
	RegX	*regx;	// now we need at least nlisany...
	Inst	program[NPROG];
	Inst	*progp;
	Inst	*startinst;	/* First inst. of program; might not be program[0] */
	Inst	*bstartinst;	/* same for backwards machine */
	Ilist	*tl, *nl;	/* This list, next list */
	Ilist	list[2][NLIST];

	#define	NSTACK	20
	Node	andstack[NSTACK];
	Node	*andp;
	int	atorstack[NSTACK];
	int	*atorp;
	int	lastwasand;	/* Last token was operand */
	int	cursubid;
	int	subidstack[NSTACK];
	int	*subidp;
	int	backwards;
	int	nbra;
	Cstrip *strip;
	Posn lowbound;	// marks the earliest position still being considered

	#define	DICODE	10	/* allocation increment */
	int	nicode;		/* number active */
	int	Nicode;		/* high water mark */
//	RCode	**icode;
	long	**icode;
	int	negateicode;
	int lastc;	// for possible initial BOL check

	Inst *	newinst(int t);
	void	addinst(Ilist *lp, Inst *inst, Rangeset *sep);
	void	pushand(Inst *f, Inst *lp);
	void	pushator(int t);
	Node	*popand(int op);
	int		popator(void);
	void	startlex(Cstrip *s);
	int		esccode(void);
	int		lex(void);
	long	nextrec(void);
	void	regxop(int t);
	void	operand(int t);
	void	evaluntil(int pri);
	void	optimize(Inst *start);
	void	bldcicode(void);
	int		icodematch(int icodeno, int c, bool negate=false);
	int		icodencmatch(int icodeno, int c, bool negate=false);
	Inst *	realcompile(Cstrip *s);
   #ifdef	RXDEBUG
	void dumpstack(void);
	void dump(void);
   #endif
};

_RegX_Internal::_RegX_Internal(RegX *r) 
{
	memset(this, 0, sizeof(_RegX_Internal));
	regx = r;	// AFTER clearing!!
	lastc = '\n'; // force BOL initially
}

_RegX_Internal::~_RegX_Internal()
{
	for(int i=0; i<nicode; i++)
		free(icode[i]);
}



void panic (char *msg) {
	printf("%s\n", msg);
	exit(11);
}


Inst *
_RegX_Internal::newinst(int t)
{
	if(progp >= &program[NPROG])
		throw RegX::Error(RegX::EXPR_TOO_LONG);
	progp->type = t;
	progp->left = 0;
	progp->right = 0;
	DPRINTF(("New Instruction %s at %d\n",
		iname(t), progp-program);)
	return progp++;
}

Inst *
_RegX_Internal::realcompile(Cstrip *s)
{
	int token;

	s->pos(0);	// ensure at start	
	startlex(s);
	atorp = atorstack;
	andp = andstack;
	subidp = subidstack;
	cursubid = 0;
	lastwasand = false;
	/* Start with a low priority operator to prime parser */
	pushator(START-1);
	while((token=lex()) != END){
		if((token&ISATOR) == OPERATOR)
			regxop(token);
		else
			operand(token);
	}
	/* Close with a low priority operator */
	evaluntil(START);
	/* Force END */
	operand(END);
	evaluntil(START);
	if(nbra)
		throw RegX::Error(RegX::LEFT_PAR_ERR);
	--andp;	/* points to first and only operand */
	return andp->first;
}

void
_RegX_Internal::operand(int t)
{
	Inst *i;
	DPRINTF(("Operand %s\n", iname(t));)
	if(lastwasand)
		regxop(CAT);	/* catenate is implicit */
	i = newinst(t);
	if(t == CICODE){
		if(negateicode)
			i->type = NCICODE;	/* UGH */
		i->ricode = nicode-1;		/* UGH */
	}
	pushand(i, i);
	lastwasand = true;
	DPRINTF(("Done Operand %s\n", iname(t));)
}

void
_RegX_Internal::regxop(int t)
{
	DPRINTF(("Operator %s\n", iname(t));)
	if(t==RBRA && --nbra<0)
		throw RegX::Error(RegX::RIGHT_PAR_ERR);
	if(t==LBRA){
		cursubid++;	/* silently ignored */
		nbra++;
		if(lastwasand)
			regxop(CAT);
	}else
		evaluntil(t);
	if(t!=RBRA)
		pushator(t);
	lastwasand = false;
	if(t==STAR || t==QUEST || t==PLUS || t==RBRA)
		lastwasand = true;	/* these look like operands */
	DPRINTF(("Done Operator %s\n", iname(t));)
}

void
cant(char *s)
{
	char buf[100];

	sprintf(buf, "regexp: can't happen: %s", s);
	panic(buf);
}

void
_RegX_Internal::pushand(Inst *f, Inst *lp)
{
	DPRINTF(("Push OpMarker first %d  last %d\n",
		 f-program, lp-program);)
	if(andp >= &andstack[NSTACK])
		cant("operand stack overflow");
	andp->first = f;
	andp->last = lp;
	andp++;
}

void
_RegX_Internal::pushator(int t)
{
	DPRINTF(("Push Operator %s\n", iname(t));)
	if(atorp >= &atorstack[NSTACK])
		cant("operator stack overflow");
	*atorp++=t;
	if(cursubid >= NSUBEXP)
		*subidp++= -1;
	else
		*subidp++=cursubid;
}

Node *
_RegX_Internal::popand(int op)
{
	if(andp <= &andstack[0]) {
		DPRINTF(("Throwing operand stack error");)
		if(op)
			throw RegX::Error(RegX::MISSING_OP, op);
		else
			throw RegX::Error(RegX::BAD_REG_EXPR);
	}
	DPRINTF(("Pop OpMarker %s",
		iname(andp[-1].first->type));)
	DPRINTF(("/%s (first %d  last %d)\n",
		iname(andp[-1].last->type),
		andp[-1].first-program,
		andp[-1].last-program);)
	return --andp;
}

int
_RegX_Internal::popator(void)
{
	DPRINTF(("Pop Operator %s\n", iname(atorp[-1]));)
	if(atorp <= &atorstack[0])
		cant("operator stack underflow");
	--subidp;
	return *--atorp;
}

void
_RegX_Internal::evaluntil(int pri)
{
	Node *op1, *op2, *t;
	Inst *inst1, *inst2;

	DPRINTF(("Eval Until %s (%x)\n", iname(pri), pri);)
	while(pri==RBRA || atorp[-1]>=pri){
		#ifdef RXDEBUG
		if (pri!=RBRA)
			printf("...passed against %s\n", iname(atorp[-1]));
		#endif
		switch(popator()){
		case LBRA:
			DPRINTF(("Eval LBRA\n");)
			op1 = popand('(');
			inst2 = newinst(RBRA);
			inst2->subid = *subidp;
			op1->last->next = inst2;
			inst1 = newinst(LBRA);
			inst1->subid = *subidp;
			inst1->next = op1->first;
			pushand(inst1, inst2);
			return;		/* must have been RBRA */
		default:
			panic("unknown regexp operator");
			break;
		case OR:
			DPRINTF(("Eval OR\n");)
			op2 = popand('|');
			op1 = popand('|');
			inst2 = newinst(NOP);
			op2->last->next = inst2;
			op1->last->next = inst2;
			inst1 = newinst(OR);
			inst1->right = op1->first;
			inst1->left = op2->first;
			pushand(inst1, inst2);
			break;
		case CAT:
			DPRINTF(("Eval CAT\n");)
			op2 = popand(0);
			op1 = popand(0);
			if(backwards && op2->first->type!=END)
				t = op1, op1 = op2, op2 = t;
			op1->last->next = op2->first;
			pushand(op1->first, op2->last);
			break;
		case STAR:
			DPRINTF(("Eval STAR\n");)
			op2 = popand('*');
			inst1 = newinst(OR);
			op2->last->next = inst1;
			inst1->right = op2->first;
			pushand(inst1, inst1);
			break;
		case PLUS:
			DPRINTF(("Eval PLUS\n");)
			op2 = popand('+');
			inst1 = newinst(OR);
			op2->last->next = inst1;
			inst1->right = op2->first;
			pushand(op2->first, inst1);
			break;
		case QUEST:
			DPRINTF(("Eval QUEST\n");)
			op2 = popand('?');
			inst1 = newinst(OR);
			inst2 = newinst(NOP);
			inst1->left = inst2;
			inst1->right = op2->first;
			op2->last->next = inst2;
			pushand(inst1, inst2);
			break;
		}
	}
	DPRINTF(("Finished Eval %s\n", iname(pri));)
}


void
_RegX_Internal::optimize(Inst *start)
{
	Inst *inst, *target;

	for(inst=start; inst->type!=END; inst++){
		target = inst->next;
		while(target->type == NOP)
			target = target->next;
		inst->next = target;
	}
}

#ifdef	RXDEBUG

char * iname(int t)
{
	switch (t) {
	case START:
		return "START";
	case RBRA:
		return "RBRA";
	case LBRA:
		return "LBRA";
	case OR:
		return "OR";
	case CAT:
		return "CAT";
	case STAR:
		return "STAR";
	case PLUS:
		return "PLUS";
	case QUEST:
		return "QUEST";
	case ANY:
		return "ANY";
	case NOP:
		return "NOP";
	case BOL:
		return "BOL";
	case EOL:
		return "EOL";
	case CICODE:
		return "CICODE";
	case NCICODE:
		return "NCICODE";
	case END:
		return "END";
	case '\n':
		return "<newline>";
	case '\t':
		return "<tab>";
	case 0:
		return "<null>";
	default:
		// CAUTION: This bit means func can't be called twice in one printf!
		if (t < 0x7F) {
			static char charname[4] = "' '";
			charname[1] = (char)t;
			return charname;
		}
		else return "Other Op";
	}
}

void _RegX_Internal::dumpstack(void){
	Node *stk;
	int *ip;

	printf("operators\n");
	for(ip = atorstack; ip<atorp; ip++)
		printf("%x\n", *ip);
	printf("operands\n");
	for(stk = andstack; stk<andp; stk++)
		printf("%x\t%x\n", stk->first->type, stk->last->type);
}

void _RegX_Internal::dump(void){
	Inst *lp = program;
	do{
		printf("%d:\t%s", lp-program, iname(lp->type));
		if (lp->left >= program)
			printf("\t%d", lp->left-program);
		else
			printf("\t[%d]", lp->left);
		if (lp->right >= program)
			printf("\t%d\n", lp->right-program);
		else
			printf("\t[%d]\n", lp->right);
	}while(lp++->type);
	printf("=========================\n");
}
#endif

void
_RegX_Internal::startlex(Cstrip *s)
{
	strip = s;
	nbra = 0;
}

int _RegX_Internal::esccode(void){	// called if backslash encountered
	int c = strip->get();
	if(c > 0) switch (c) {
		case 'n': c='\n'; break;	// newline
		case 't': c='\t'; break;	// tab
		case 'r': c='\r'; break;	// carriage-return
		case 'b': c='\b'; break;	// backspace
		case 'f': c='\f'; break;	// formfeed
		case 'v': c='\v'; break;	// vertical tab
		case 'a': c='\a'; break;	// 'bell' (alert) char
		case 'q': c='\''; break;	// single-quote
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
				{
					DPRINTF(("first octal entry char %c\n", c);)
					int tc = c;
					c -= '0';
					for (int i=3 ; i; i--) {
						tc = strip->look();
						DPRINTF(("octal entry char %c\n", tc);)
						if (tc < '0' || tc > '7') break;
						c = (c << 3) + tc - '0';
						strip->get();
					}
					c &= 0xFF;	// octal is restricted to a byte
					DPRINTF(("value=%x next char %c\n", c, strip->look());)
				}
				break;
		case 'x':
		case 'X':
				{
					c = 0;
					for (int i=4 ; i; i--) {
						int tc = strip->look();
						DPRINTF(("hex entry char %c\n", tc);)
						if (tc >= '0' && tc <= '9') tc -= '0';
						else if (tc >= 'A' && tc <= 'F') tc += 0xA -'A';
						else if (tc >= 'a' && tc <= 'f') tc += 0xA -'a';
						else break;
						c = (c << 4) + tc;
						strip->get();
					}
					DPRINTF(("value=%x next char %c\n", c, strip->look());)
				}
				break;
	}
	return c;
}

int
_RegX_Internal::lex(void){
	int c= strip->get();
	DPRINTF(("read char '%c' %x\n", c, c);)

	switch(c){
	case '\\':
		c = esccode();
		break;
	case 0:
	case -1:
		c = END;
		strip->pos(strip->pos() - 1);	/* In case we come here again */
		break;
	case '*':
		c = STAR;
		break;
	case '?':
		c = QUEST;
		break;
	case '+':
		c = PLUS;
		break;
	case '|':
		c = OR;
		break;
	case '.':
		c = ANY;
		break;
	case '(':
		c = LBRA;
		break;
	case ')':
		c = RBRA;
		break;
	case '^':
		c = BOL;
		break;
	case '$':
		c = EOL;
		break;
	case '[':
		c = CICODE;
		bldcicode();
		break;
	}
	return c;
}

long _RegX_Internal::nextrec(void){
	int tc0 = strip->get();
	int tc1 = strip->look();
	if(tc0<=0 || (tc0=='\\' && tc1<=0))
		throw RegX::Error(RegX::BAD_CODE);
	if(tc0 == '\\'){
		return esccode() | 0x10000;	// the mark prevents match of ']' or '-' below
	}
	return tc0;	// already bumped...
}

void
_RegX_Internal::bldcicode(void)
{
	long c1, c2, n, na;
//	RCode *icodep;
	long *icodep;

	icodep = (long *)malloc(DICODE*sizeof(long));
	n = 0;
	na = DICODE;
	/* we have already seen the '[' */
	if(strip->look() == '^'){
		if (!regx->nlisany) icodep[n++] = '\n';	/* newline match in negate case dependent */
		negateicode = true;
		strip->get();
	}else
		negateicode = false;
	while((c1 = nextrec()) != ']'){
		if(c1 == '-'){
    badExp:
			free(icodep);
			throw RegX::Error(RegX::BAD_CODE);
		}
		if(n+4 >= na){		/* 3 runes plus NUL */
			na += DICODE;
			icodep = (long *)realloc(icodep, na*sizeof(long));
		}
		if(strip->look() == '-'){
			strip->get();	/* eat '-' */
			if((c2 = nextrec()) == ']')
				goto badExp;
//			icodep[n+0] = 0xFFFF;
			icodep[n+0] = 0x10000;
			icodep[n+1] = c1 & 0xFFFF;	// truncates to short (removing any mark)
			icodep[n+2] = c2 & 0xFFFF;
			n += 3;
		}else
			icodep[n++] = c1 & 0xFFFF;
	}
//	icodep[n] = 0;
	icodep[n] = -1;	// terminator
	if(nicode == Nicode){
		Nicode += DICODE;
//		icode = (RCode **)realloc(icode, Nicode*sizeof(RCode*));
		icode = (long **)realloc(icode, Nicode*sizeof(long*));
	}
	icode[nicode++] = icodep;
}

int
_RegX_Internal::icodematch(int icodeno, int c, bool negate)
{
//	RCode *p;
	long *p;

	p = icode[icodeno];
//	while(*p){
	while(*p != -1){
//		if(*p == 0xFFFF){
		if(*p == 0x10000){
			if(p[1]<=c && c<=p[2])
				return !negate;
			p += 3;
		}else if(*p++ == c)
			return !negate;
	}
	return negate;
}

int
_RegX_Internal::icodencmatch(int icodeno, int c, bool negate)
{
//	RCode *p;
	long *p;
	c = toupper(c);

	p = icode[icodeno];
//	while(*p){
	while(*p != -1){
//		if(*p == 0xFFFF){
		if(*p == 0x10000){
			if(toupper(p[1])<=c && c<=toupper(p[2]))
				return !negate;
			p += 3;
		}else if(toupper(*p++) == c)
			return !negate;
	}
	return negate;
}

/*
 * Note optimization in addinst:
 * 	*lp must be pending when addinst called; if *lp has been looked
 *		at already, the optimization is a bug.
 */
void
_RegX_Internal::addinst(Ilist *lp, Inst *inst, Rangeset *sep)
{
	Ilist *p;

	for(p = lp; p->inst; p++){
		if(p->inst==inst){
			if((sep)->p[0].p1 < p->se.p[0].p1)
				p->se= *sep;	/* this would be bug */
			return;	/* It's already there */
		}
	}
	DPRINTF(("AddInst to list %p: %s, range %d %d\n",
			 lp, iname(inst->type),
			 sep->p[0].p1, sep->p[0].p2);)
	p->inst = inst;
	p->se= *sep;
	(p+1)->inst = 0;
	if (sep->p[0].p1 < lowbound) lowbound = sep->p[0].p1;
}

/////////////////////////////////////////////////////////

RegX::RegX() : shortest(false), nlisany(false),
				 eotiseol(true), nocase(false)
{
	rx = new _RegX_Internal(this);
}

RegX::RegX(Cstrip *s) : shortest(false), nlisany(false),
				 eotiseol(true), nocase(false)
{
	rx = new _RegX_Internal(this);
	compile(s);
}

RegX::~RegX() {
	delete rx;
}

void RegX::compile(Cstrip *s)
{
	int i;
	Inst *oprogp;

	for(i=0; i<rx->nicode; i++)
		free(rx->icode[i]);
	rx->nicode = 0;
	rx->progp = rx->program;
	rx->backwards = false;
	rx->startinst = rx->realcompile(s);
	rx->optimize(rx->program);
	oprogp = rx->progp;
	rx->backwards = true;
	rx->bstartinst = rx->realcompile(s);
	rx->optimize(oprogp);
  #ifdef	RXDEBUG
	rx->dump();
	printf("start point = %d\n", rx->startinst - rx->program);
   #endif
}

void RegX::newmatch(Rangeset *sp)
{
	int i;
	DPRINTF(("New Match %d %d (prev %d %d)\n",
	 sp->p[0].p1, sp->p[0].p2,
	 sel.p[0].p1, sel.p[0].p2);)

	if(sel.p[0].p1<0 || sp->p[0].p1<sel.p[0].p1 ||
	   (sp->p[0].p1==sel.p[0].p1 && sp->p[0].p2>sel.p[0].p2)) {
		for(i = 0; i<NSUBEXP; i++)
			sel.p[i] = sp->p[i];
		rx->lowbound = sel.p[0].p1;	// ensure not too high
	}
}

void RegX::bnewmatch(Rangeset *sp)
{
        int  i;
        if(sel.p[0].p1<0 || sp->p[0].p1>sel.p[0].p2 || (sp->p[0].p1==sel.p[0].p2 && sp->p[0].p2<sel.p[0].p1))
                for(i = 0; i<NSUBEXP; i++){       /* note the reversal; p1<=p2 */
                        sel.p[i].p1 = sp->p[i].p2;
                        sel.p[i].p2 = sp->p[i].p1;
                }
}


bool RegX::scan(Cstrip * text, Posn startp, Posn eof)
{
	Rangeset sempty;	// was static... (temp but might clash in multitask)
	memset(&sempty, 0, sizeof(sempty));	// but now we need this
	int flag = 0;
	Inst *inst;
	Ilist *tlp;
	Posn p = startp;
	int nnl = 0, ntl;
	int c = 0;
	int wrapped = 0;
	int startchar = rx->startinst->type<OPERATOR? rx->startinst->type : 0;
	DPRINTF(("starting scan with strip %p pos %d eof %d\n", text, startp, eof);)

	if (startp == 0) c = '\n';	// force BOL at BOT!
	else {	// get prev char ready for lastc
		if (text->pos(startp-1)) // temporary back up (a bit slow but necessary)
			c = text->get();	// should move back to startp
		else throw Error(SCAN_FAIL);
	}	
	if (text->pos() != startp && !text->pos(startp)) return false;

	rx->list[0][0].inst = rx->list[1][0].inst = 0;
	DPRINTF(("lastc will be '%c'(%x)\n", c, c);)
	sel.p[0].p1 = -1;
	rx->lowbound = startp;
	/* Execute machine once for each character */
	for(;;p++){
	doloop:
		rx->lastc = c;	// tracking for BOL
		DPRINTF(("do loop -- lastc set to'%c'(%x)\n", rx->lastc, rx->lastc);)
		// Duplicated here (99:4:29) to report match faster...:
		if(sel.p[0].p1>=0 && nnl==0)
			break;
		c = text->get();
		DPRINTF(("+++ appraising char '%c' %x +++\n", c, c);)
		DPRINTF(("low bound at %d\n", rx->lowbound);)
		text->lowBound(rx->lowbound);	// tell strip where we are
		if (sel.p[0].p1 < 0)	// no current match
			rx->lowbound = p;	// may get moved back more
		if(p>=eof || c<0){	
			switch(wrapped++){
			case 0:		/* let loop run one more click */
			case 2:
				DPRINTF(("extra click at end of text\n");)
				break;
			case 1:		/* expired; wrap to beginning */
			#ifdef ALLOW_WRAP
			// we never wrap in this version
				if(sel.p[0].p1>=0 || eof!=INFINITY)
					goto Return;
				rx->list[0][0].inst = rx->list[1][0].inst = 0;
				text->pos(0);
				p = 0;
				goto doloop;
			#endif
			default:
				goto Return;
			}
		}
		else if(((wrapped && p>=startp) || sel.p[0].p1>=0) && nnl==0)
			break;
		/* fast check for first char */
		if (nocase) {
			if (startchar && nnl==0
				 && toupper(c) != toupper(startchar))
			continue;
		} else {
			if(startchar && nnl==0 && c!=startchar)
			continue;
		}
		rx->tl = rx->list[flag];
		rx->nl = rx->list[flag^=1];
		rx->nl->inst = 0;
		ntl = nnl;
		nnl = 0;
		if(sel.p[0].p1<0 && (!wrapped || p<startp || startp==eof)){
			/* Add first instruction to this list */
			if(++ntl >= NLIST)
	Overflow:
			throw Error(OVERFLOW);
			sempty.p[0].p1 = p;
			rx->addinst(rx->tl, rx->startinst, &sempty);
		}
		/* Execute machine until this list is empty */
		for(tlp = rx->tl; inst = tlp->inst; tlp++) {	/* assignment = */
	Switchstmt:
			DPRINTF((" Executing Inst %s\n", iname(inst->type));)
			switch(inst->type){
			default:	/* regular character */
				if (inst->type==c ||
				 (nocase && toupper(inst->type)==toupper(c)) ){
	Addinst:
					if(++nnl >= NLIST)
						goto Overflow;
					rx->addinst(rx->nl, inst->next, &tlp->se);
				}
				break;
			case LBRA:
				if(inst->subid>=0)
					tlp->se.p[inst->subid].p1 = p;
				inst = inst->next;
				goto Switchstmt;
			case RBRA:
				if(inst->subid>=0)
					tlp->se.p[inst->subid].p2 = p;
				inst = inst->next;
				goto Switchstmt;
			case ANY:
				if(c!='\n' || nlisany)
					goto Addinst;
				break;
			case BOL:
				DPRINTF(("BOL test p=%d  curr='%c' last='%c'(%x)\n",
					 p, c, rx->lastc, rx->lastc);)
				if(rx->lastc =='\n') {
						DPRINTF((" last = \\n -- stepping\n");)
	Step:
					inst = inst->next;
					DPRINTF(("did Step\n");)
					goto Switchstmt;
				}
				break;
			case EOL:
				// (note -- if eof (Posn param) is set, it does *not* match eol)
				if(c == '\n' || (c<0 && eotiseol))
					goto Step;
				break;
			case NOP:	// in case optimize missed it...
						// (Added for experiments -- not really needed)
					goto Step;
				break;
			case CICODE:
				if (nocase) {
					if(c>=0 && rx->icodencmatch(inst->ricode, c))
						goto Addinst;
				} else {
					if(c>=0 && rx->icodematch(inst->ricode, c))
						goto Addinst;
				}
				break;
			case NCICODE:
				if (nocase) {
					if(c>=0 && rx->icodencmatch(inst->ricode, c, true))
						goto Addinst;
				} else {
					if(c>=0 && rx->icodematch(inst->ricode, c, true))
						goto Addinst;
				}
				break;
			case OR:
				/* evaluate right choice later */
				if(++ntl >= NLIST)
					goto Overflow;
				rx->addinst(tlp, inst->right, &tlp->se);
				/* efficiency: advance and re-evaluate */
				inst = inst->left;
				goto Switchstmt;
			case END:	/* Match! */
				tlp->se.p[0].p2 = p;
				DPRINTF(("END -- curr is %c(%x)\n", c, c);)
				newmatch(&tlp->se);
				if (shortest) goto Return;	// this mode accepts first match
				break;
			}
		}
	}
    Return:
	DPRINTF(("scan returning -- last pos scanned %d\n", text->pos()-1);)
    return (sel.p[0].p1>=0);
}

//////////////////////////////////////////////////////////////////////
// This part of the Bell Labs code has not been updated to function
// with Cstrips, because these don't handle backward scanning yet.
// (Maybe sometime in the future, but at the moment they are oriented
// toward non-seekable streams).
// The code will perhaps still work on a simple array of (16-bit in
// this case) characters, but the features added to forward scanning
// have not been put into this code yet

bool RegX::scanback(unsigned short * text, Posn startp)
{
	Rangeset sempty;	// was static... (temp but might clash in multitask)
	memset(&sempty, 0, sizeof(sempty));	// but now we need this
	int flag = 0;
	Inst *inst;
	Ilist *tlp;
	Posn p = startp;
	int nnl = 0, ntl;
	int c = 0;
	int wrapped = 0;
	int startchar =
		 rx->bstartinst->type<OPERATOR? rx->bstartinst->type : 0;
	unsigned short * blkp = text+startp;
	 rx->lastc = 0;

	rx->list[0][0].inst = rx->list[1][0].inst = 0;
	sel.p[0].p1= -1;
	/* Execute machine once for each character, including terminal NUL */
	for(;;--p){
	doloop:
		rx->lastc = c;
		if(p <= 0 || (c = *blkp--)==-1){
			#ifdef ALLOW_WRAP
			// this version does not use wraparound
			switch(wrapped++){
			case 0:		/* let loop run one more click */
			case 2:
				break;
			case 1:		/* expired; wrap to end */
				if(sel.p[0].p1>=0)
			case 3:
					goto Return;
				rx->list[0][0].inst = rx->list[1][0].inst = 0;
				for (p=0; text[p]; p++){/*loop*/}
				blkp = text+p;
				goto doloop;
			default:
				goto Return;
			}
			#endif
			goto Return;
		}else if(((wrapped && p<=startp) || sel.p[0].p1>0) && nnl==0)
			break;
		/* fast check for first char */
		if(startchar && nnl==0 && c!=startchar)
			continue;
		rx->tl = rx->list[flag];
		rx->nl = rx->list[flag^=1];
		rx->nl->inst = 0;
		ntl = nnl;
		nnl = 0;
		if(sel.p[0].p1<0 && (!wrapped || p>startp)){
			/* Add first instruction to this list */
			if(++ntl >= NLIST)
	Overflow:
			throw Error(OVERFLOW);
			/* the minus is so the optimizations in addinst work */
			sempty.p[0].p1 = -p;
			rx->addinst(rx->tl, rx->bstartinst, &sempty);
		}
		/* Execute machine until this list is empty */
		for(tlp = rx->tl; inst = tlp->inst; tlp++){	/* assignment = */
	Switchstmt:
			switch(inst->type){
			default:	/* regular character */
				if(inst->type == c){
	Addinst:
					if(++nnl >= NLIST)
						goto Overflow;
					rx->addinst(rx->nl, inst->next, &tlp->se);
				}
				break;
			case LBRA:
				if(inst->subid>=0)
					tlp->se.p[inst->subid].p1 = p;
				inst = inst->next;
				goto Switchstmt;
			case RBRA:
				if(inst->subid >= 0)
					tlp->se.p[inst->subid].p2 = p;
				inst = inst->next;
				goto Switchstmt;
			case ANY:
				if(c != '\n')
					goto Addinst;
				break;
			case BOL:
				if(c=='\n' || p==0){
	Step:
					inst = inst->next;
					goto Switchstmt;
				}
				break;
			case EOL:
				//back scan doesn't take EOT as EOL for now...
				if(rx->lastc =='\n') {
						goto Step;
				}
				#ifdef SAMs_WAY_NOT_USED
				if(f->getci<f->ngetc-1){
					if(f->getcbuf[f->getci+1]=='\n')
						goto Step;
				}else if(p<f->nrunes-1){
					unsigned short c;
					if(Fchars(f, &c, p, p+1)==1 && c=='\n')
						goto Step;
				}
				#endif
				break;
			case CICODE:
				if(c>=0 && rx->icodematch(inst->ricode, c))
					goto Addinst;
				break;
			case NCICODE:
				if(c>=0 && rx->icodematch(inst->ricode, c, true))
					goto Addinst;
				break;
			case OR:
				/* evaluate right choice later */
				if(++ntl >= NLIST)
					goto Overflow;
				rx->addinst(tlp, inst->right, &tlp->se);
				/* efficiency: advance and re-evaluate */
				inst = inst->left;
				goto Switchstmt;
			case END:	/* Match! */
				tlp->se.p[0].p1 = -tlp->se.p[0].p1; /* minus sign */
				tlp->se.p[0].p2 = p;
				bnewmatch(&tlp->se);
				break;
			}
		}
	}
    Return:
	return sel.p[0].p1>=0;
}

/////////////////////////////////////////////////////////////
