//////////////////////////////////////////////////////////////////
/*! \file
 'RegX' "Structural Regular Expression" Matching Class
 -- Unicode & Multiline aware.
 Copyright (c) 1999-2003 Peter J. Goodeve
*/

//////////////////////////////////////////////////////////////////

#include <sys/types.h>

class Cstrip;

typedef unsigned short RCode;
#define RCODESIZE sizeof(RCode)
#define NSUBEXP	10
#ifndef INFINITY
  #define	INFINITY	0x7FFFFFFFL
#endif

/*! file position  (from 0) */
typedef long		Posn;		

/*! Structure to represent a range of characters within the text */
struct Range
{
	Posn	p1, p2;
};

/*! Array of ranges to represent matched subsegments */
struct Rangeset
{
	Range	p[NSUBEXP];
};


/////////////////////////////////////////////////////////////////////////

//! 'RegX' "Structural Regular Expression" Matching Class
/*!	The regular expressions managed by this class are almost identical
	in form to those familiar from grep and so on, but rather than matching
	a complete line in the scanned text they return the matched text segment
	itself.  The match may span several lines or just part of one.
	
	Both the regular expression to be compiled and the text to be scanned
	are presented to the class as 'CStrips' that deliver characters (perhaps
	originally UTF-8) as 16-bit values (actually signed integers, to handle
	EOF and errors) on demand. (see cstrip.h et al.)
	The actual sources will be classes such as UtfStrip or FileStrip derived
	from Cstrip itself.
	
	This public class is actually just a container for the structure in regx.cpp
	itself that does the real work.
*/
class RegX {
	struct _RegX_Internal *rx;	// most of this would just get in the way
  protected:
	//! Fills 'sel' with the final results of the match 
	void newmatch(Rangeset *sp);
	//! Fills 'sel' with the final results of a backward match [unused!] 
	void bnewmatch(Rangeset *sp);
  public:
	//! The results of a match (begin, end, and subsegments) are returned here
	Rangeset	sel;
	bool
			//! set this true to get shortest match (default is longest)
			shortest,
			//! set true for '.' to match newline (default false)
			nlisany,	
			//! (default) true for '$' to match end-of-text without newline
			eotiseol,	
			//! set true for case-insensitive matching (default false)
			nocase;		
	RegX();
	//! Constructor: parameter 's' is the (initial) regular expression to be used.
	RegX(Cstrip *s);
	~RegX();
	enum {EXPR_TOO_LONG, LEFT_PAR_ERR, RIGHT_PAR_ERR,
		 MISSING_OP, BAD_REG_EXPR, BAD_CODE, OVERFLOW, SCAN_FAIL};
	//! Exception class thrown by compile and scan errors
	struct Error {
		int code;
		char c;
		Error(int i=0, char ec='\0') : code(i), c(ec) {}
	};
	//! Compile regular expression 's' into state-machine for matching
	void compile(Cstrip *s);
	//! scan 'text' for a match.
	/*!	The scan begins at character position 'startp'.
		If 'eof' is supplied, the scan will end at that point, rather than
		the actual end-of-text.
	*/
	bool scan(Cstrip *text, Posn startp=0, Posn eof=INFINITY);
	//! not implemented for Cstrip yet: 
	bool scanback(unsigned short * text, Posn startp);	// can't scan strip backward (yet)
};


